# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Lifecycle class for craft-parts.

PENDING DEPRECATION: we're moving this to a craft-application LifecycleService
"""
import os
import pathlib
import shlex
from typing import Any, Dict, List

from craft_cli import CraftError, emit
from craft_parts import LifecycleManager, PartsError, Step
from xdg import BaseDirectory  # type: ignore[import]

from charmcraft import charm_builder, instrum


class PartsLifecycle:
    """Create and manage the parts lifecycle.

    :param all_parts: A dictionary containing the parts defined in the project.
    :param work_dir: The working directory for parts processing.
    :param project_dir: The directory containing the charm project.
    :param ignore_local_sources: A list of local source patterns to be ignored.
    :param name: Charm name as defined in ``metadata.yaml``.
    """

    def __init__(
        self,
        all_parts: Dict[str, Any],
        *,
        work_dir: pathlib.Path,
        project_dir: pathlib.Path,
        project_name: str,
        ignore_local_sources: List[str],
    ):
        self._all_parts = all_parts.copy()
        self._project_dir = project_dir

        # set the cache dir for parts package management
        cache_dir = BaseDirectory.save_cache_path("charmcraft")

        try:
            self._lcm = LifecycleManager(
                {"parts": all_parts},
                application_name="charmcraft",
                work_dir=work_dir,
                cache_dir=cache_dir,
                ignore_local_sources=ignore_local_sources,
                project_name=project_name,
            )
        except PartsError as err:
            raise CraftError(f"Error bootstrapping lifecycle manager: {err}") from err

    @property
    def prime_dir(self) -> pathlib.Path:
        """Return the parts prime directory path."""
        return self._lcm.project_info.prime_dir

    def run(self, target_step: Step) -> None:
        """Run the parts lifecycle.

        :param target_step: The final step to execute.

        :raises CraftError: On error during lifecycle ops.
        :raises RuntimeError: On unexpected error.
        """
        previous_dir = os.getcwd()
        try:
            os.chdir(self._project_dir)

            # invalidate build if packing a charm and entrypoint changed
            if "charm" in self._all_parts:
                charm_part = self._all_parts["charm"]
                if charm_part.get("plugin") == "charm":
                    entrypoint = os.path.normpath(charm_part["charm-entrypoint"])
                    dis_entrypoint = os.path.normpath(_get_dispatch_entrypoint(self.prime_dir))
                    if entrypoint != dis_entrypoint:
                        self._lcm.clean(Step.BUILD, part_names=["charm"])
                        self._lcm.reload_state()

            emit.debug(f"Executing parts lifecycle in {str(self._project_dir)!r}")
            actions = self._lcm.plan(target_step)
            emit.debug(f"Parts actions: {actions}")
            with instrum.Timer("Running action executor") as executor_timer:
                with self._lcm.action_executor() as aex:
                    executor_timer.mark("Context enter")
                    for act in actions:
                        emit.progress(f"Running step {act.step.name} for part {act.part_name!r}")
                        with instrum.Timer("Running step", step=act.step.name, part=act.part_name):  # type: ignore[arg-type]
                            with emit.open_stream("Execute action") as stream:
                                aex.execute([act], stdout=stream, stderr=stream)
                    executor_timer.mark("Context exit")

        except RuntimeError as err:
            raise RuntimeError(f"Parts processing internal error: {err}") from err
        except OSError as err:
            msg = err.strerror
            if err.filename:
                msg = f"{err.filename}: {msg}"
            raise CraftError(f"Parts processing error: {msg}") from err
        except Exception as err:
            raise CraftError(f"Parts processing error: {err}") from err
        finally:
            os.chdir(previous_dir)


def _get_dispatch_entrypoint(dirname: pathlib.Path) -> str:
    """Read the entrypoint from the dispatch file."""
    dispatch = dirname / charm_builder.DISPATCH_FILENAME
    entrypoint_str = ""
    try:
        with dispatch.open("rt", encoding="utf8") as fh:
            last_line = None
            for line in fh:
                if line.strip():
                    last_line = line
            if last_line:
                entrypoint_str = shlex.split(last_line)[-1]
    except (OSError, UnicodeDecodeError):
        return ""

    return entrypoint_str



# Copyright 2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import pathlib

import pytest

from charmcraft import env


def test_get_managed_environment_home_path():
    dirpath = env.get_managed_environment_home_path()

    assert dirpath == pathlib.Path("/root")


def test_get_managed_environment_log_path():
    dirpath = env.get_managed_environment_log_path()

    assert dirpath == pathlib.Path("/tmp/charmcraft.log")


def test_get_managed_environment_project_path():
    dirpath = env.get_managed_environment_project_path()

    assert dirpath == pathlib.Path("/root/project")


def test_get_managed_environment_snap_channel_none(monkeypatch):
    monkeypatch.delenv("CHARMCRAFT_INSTALL_SNAP_CHANNEL", raising=False)

    assert env.get_managed_environment_snap_channel() is None


def test_get_managed_environment_snap_channel(monkeypatch):
    monkeypatch.setenv("CHARMCRAFT_INSTALL_SNAP_CHANNEL", "latest/edge")

    assert env.get_managed_environment_snap_channel() == "latest/edge"


@pytest.mark.parametrize(
    ("snap_name", "snap", "result"),
    [
        (None, None, False),
        (None, "/snap/charmcraft/x1", False),
        ("charmcraft", None, False),
        ("charmcraft", "/snap/charmcraft/x1", True),
    ],
)
def test_is_charmcraft_running_from_snap(monkeypatch, snap_name, snap, result):
    if snap_name is None:
        monkeypatch.delenv("SNAP_NAME", raising=False)
    else:
        monkeypatch.setenv("SNAP_NAME", snap_name)

    if snap is None:
        monkeypatch.delenv("SNAP", raising=False)
    else:
        monkeypatch.setenv("SNAP", snap)

    assert env.is_charmcraft_running_from_snap() == result


@pytest.mark.parametrize(
    ("developer", "result"),
    [
        (None, False),
        ("y", True),
        ("n", False),
        ("Y", True),
        ("N", False),
        ("true", True),
        ("false", False),
        ("TRUE", True),
        ("FALSE", False),
        ("yes", True),
        ("no", False),
        ("YES", True),
        ("NO", False),
        ("1", True),
        ("0", False),
    ],
)
def test_is_charmcraft_running_in_developer_mode(monkeypatch, developer, result):
    if developer is None:
        monkeypatch.delenv("CHARMCRAFT_DEVELOPER", raising=False)
    else:
        monkeypatch.setenv("CHARMCRAFT_DEVELOPER", developer)

    assert env.is_charmcraft_running_in_developer_mode() == result


@pytest.mark.parametrize(
    ("managed", "result"),
    [
        (None, False),
        ("y", True),
        ("n", False),
        ("1", True),
        ("0", False),
    ],
)
def test_is_charmcraft_running_in_managed_mode(monkeypatch, managed, result):
    if managed is None:
        monkeypatch.delenv("CHARMCRAFT_MANAGED_MODE", raising=False)
    else:
        monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", managed)

    assert env.is_charmcraft_running_in_managed_mode() == result



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Bundle plugin for craft-parts."""
import sys
from typing import Any, Dict, List, Set

from craft_parts import plugins


class BundlePluginProperties(plugins.PluginProperties, plugins.PluginModel):
    """Properties used to pack bundles."""

    source: str

    @classmethod
    def unmarshal(cls, data: Dict[str, Any]):
        """Populate bundle properties from the part specification.

        :param data: A dictionary containing part properties.

        :return: The populated plugin properties data object.

        :raise pydantic.ValidationError: If validation fails.
        """
        plugin_data = plugins.extract_plugin_properties(
            data, plugin_name="bundle", required=["source"]
        )
        return cls(**plugin_data)


class BundlePlugin(plugins.Plugin):
    """Prepare a bundle for packing.

    Extra files to be included in the bundle payload must be listed under
    the ``prime`` file filter.
    """

    properties_class = BundlePluginProperties

    @classmethod
    def get_build_snaps(cls) -> Set[str]:
        """Return a set of required snaps to install in the build environment."""
        return set()

    def get_build_packages(self) -> Set[str]:
        """Return a set of required packages to install in the build environment."""
        return set()

    def get_build_environment(self) -> Dict[str, str]:
        """Return a dictionary with the environment to use in the build step."""
        return {}

    def get_build_commands(self) -> List[str]:
        """Return a list of commands to run during the build step."""
        install_dir = self._part_info.part_install_dir
        if sys.platform == "linux":
            cp_cmd = "cp --archive --link --no-dereference"
        else:
            cp_cmd = "cp -R -p -P"

        return [
            f'mkdir -p "{install_dir}"',
            f'{cp_cmd} * "{install_dir}"',
        ]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'extensions' command."""
import argparse
from textwrap import dedent
from typing import Dict

import tabulate
import yaml
from craft_cli import emit

from charmcraft import extensions
from charmcraft.cmdbase import BaseCommand
from charmcraft.models.extension import ExtensionModel

_overview = """
Extensions allow user use pre-defined configuration in their charmcraft.yaml.
"""


class ListExtensionsCommand(BaseCommand):
    """List available extensions for all supported bases."""

    common = True
    name = "list-extensions"
    help_msg = "List available extensions for all supported bases"
    overview = dedent(
        """
        List available extensions and their corresponding bases.
        """
    )

    def run(self, parsed_args: argparse.Namespace):
        """Print the list of available extensions and their bases."""
        extension_presentation: Dict[str, ExtensionModel] = {}

        for extension_name in extensions.registry.get_extension_names():
            extension_class = extensions.registry.get_extension_class(extension_name)
            extension_bases = list(extension_class.get_supported_bases())
            extension_presentation[extension_name] = ExtensionModel(
                name=extension_name, bases=extension_bases
            )

        printable_extensions = sorted(
            [v.marshal() for v in extension_presentation.values()],
            key=lambda d: d["Extension name"],
        )
        emit.message(tabulate.tabulate(printable_extensions, headers="keys"))


class ExtensionsCommand(ListExtensionsCommand):
    """A command alias to list the available extensions."""

    common = False
    name = "extensions"
    hidden = True

    # Workaround for AST test that cannot see run()
    def run(self, parsed_args: argparse.Namespace):
        """Print the list of available extensions and their bases."""
        ListExtensionsCommand.run(self, parsed_args)


class ExpandExtensionsCommand(BaseCommand):
    """Expand the extensions in the charmcraft.yaml file."""

    common = True
    name = "expand-extensions"
    help_msg = "Expand extensions in charmcraft.yaml"
    overview = dedent(
        """
        Expand charmcraft.yaml using the extensions specified in the file and
        output the resulting configuration to the terminal.

        This allows you to see how the extensions used modify your existing
        charmcraft.yaml file.
        """
    )

    def run(self, parsed_args: argparse.Namespace):
        """Print the project's specification with the extensions expanded."""
        self._check_config(config_file=True)
        emit.message(
            yaml.dump(
                self.config.dict(
                    by_alias=True, exclude={"project", "metadata_legacy"}, exclude_none=True
                )
            )
        )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Handlers for metadata.yaml file."""

import contextlib
import logging
import pathlib
import shutil
from typing import TYPE_CHECKING, Any, Dict

import pydantic
import yaml
from craft_cli import CraftError, emit

from charmcraft.const import (
    CHARM_METADATA_KEYS,
    CHARM_METADATA_KEYS_ALIAS,
    METADATA_FILENAME,
)
from charmcraft.format import format_pydantic_errors
from charmcraft.models.metadata import BundleMetadataLegacy, CharmMetadataLegacy

if TYPE_CHECKING:
    from charmcraft.models.charmcraft import CharmcraftConfig

logger = logging.getLogger(__name__)


def read_metadata_yaml(charm_dir: pathlib.Path) -> Dict[str, Any]:
    """Parse project's metadata.yaml.

    :returns: the YAML decoded metadata.yaml content
    """
    metadata_path = charm_dir / METADATA_FILENAME
    emit.debug(f"Reading {str(metadata_path)!r}")
    with metadata_path.open("rt", encoding="utf8") as fh:
        return yaml.safe_load(fh)


def parse_charm_metadata_yaml(
    charm_dir: pathlib.Path, allow_basic: bool = False
) -> CharmMetadataLegacy:
    """Parse project's legacy metadata.yaml that used for charms.

    :returns: a CharmMetadataLegacy object.

    :raises: CraftError if metadata.yaml does not exist or is not valid.
    """
    try:
        metadata = read_metadata_yaml(charm_dir)
    except OSError as exc:
        raise CraftError(f"Cannot read the metadata.yaml file: {exc!r}") from exc
    if not isinstance(metadata, dict):
        raise CraftError(f"The {charm_dir / METADATA_FILENAME} file is not valid YAML.")

    emit.debug("Validating metadata keys")
    try:
        return CharmMetadataLegacy.unmarshal(metadata)
    except pydantic.ValidationError as error:
        if allow_basic:
            emit.progress(
                format_pydantic_errors(error.errors(), file_name=METADATA_FILENAME), permanent=True
            )
            emit.debug("Falling back to basic metadata.yaml")
            metadata_basic = {
                k: v for k, v in metadata.items() if k in ("name", "summary", "description")
            }
            return CharmMetadataLegacy.unmarshal(metadata_basic)
        raise


def parse_bundle_metadata_yaml(charm_dir: pathlib.Path) -> BundleMetadataLegacy:
    """Parse project's legacy metadata.yaml that used for bundles.

    :returns: a BundleMetadataLegacy object.

    :raises: CraftError if metadata.yaml does not exist or is not valid.
    """
    try:
        metadata = read_metadata_yaml(charm_dir)
    except OSError as exc:
        raise CraftError(f"Cannot read the metadata.yaml file: {exc!r}") from exc
    if not isinstance(metadata, dict):
        raise CraftError(f"The {charm_dir / METADATA_FILENAME} file is not valid YAML.")

    emit.debug("Validating metadata keys")
    return BundleMetadataLegacy.unmarshal(metadata)


def create_metadata_yaml(
    charm_dir: pathlib.Path,
    charmcraft_config: "CharmcraftConfig",
) -> pathlib.Path:
    """Create metadata.yaml in charm_dir for given project configuration.

    Use CHARM_METADATA_KEYS and CHARM_METADATA_KEYS_ALIAS to filter the keys.

    :param charm_dir: Directory to create Charm in.
    :param charmcraft_config: Charmcraft configuration object.

    :returns: Path to created metadata.yaml.
    """
    original_file_path = charmcraft_config.project.dirpath / METADATA_FILENAME
    target_file_path = charm_dir / METADATA_FILENAME

    # Copy metadata.yaml if it exists, otherwise create it from CharmcraftConfig.
    if original_file_path.exists():
        # In the build / test process, the original file may be the same as the target file.
        with contextlib.suppress(shutil.SameFileError):
            shutil.copyfile(original_file_path, target_file_path)
    else:
        # metadata.yaml not exists, create it from config
        metadata = charmcraft_config.dict(
            include=CHARM_METADATA_KEYS.union(CHARM_METADATA_KEYS_ALIAS),
            exclude_none=True,
            by_alias=True,
        )

        # convert to legacy metadata format
        if (title := metadata.pop("title", None)) is not None:
            metadata["display-name"] = title

        if (links := metadata.pop("links", None)) is not None:
            if (documentation := links.pop("documentation", None)) is not None:
                metadata["docs"] = documentation
            if (issues := links.pop("issues", None)) is not None:
                metadata["issues"] = issues
            if (contact := links.pop("contact", None)) is not None:
                metadata["maintainers"] = contact
            if (source := links.pop("source", None)) is not None:
                metadata["source"] = source
            if (website := links.pop("website", None)) is not None:
                metadata["website"] = website

        target_file_path.write_text(yaml.dump(metadata))

    return target_file_path



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import datetime
import json
from unittest.mock import patch

import pytest
import yaml
from craft_cli import CraftError

from charmcraft import __version__, linters
from charmcraft.const import IMAGE_INFO_ENV_VAR
from charmcraft.metafiles.manifest import create_manifest
from charmcraft.models.charmcraft import Base, BasesConfiguration
from charmcraft.utils import OSPlatform


def test_manifest_simple_ok(tmp_path):
    """Simple construct."""
    bases_config = BasesConfiguration(
        **{
            "build-on": [
                Base(
                    name="test-name",
                    channel="test-channel",
                ),
            ],
            "run-on": [
                Base(
                    name="test-name",
                    channel="test-channel",
                    architectures=["arch1"],
                ),
                Base(
                    name="test-name2",
                    channel="test-channel2",
                    architectures=["arch1", "arch2"],
                ),
            ],
        }
    )

    linting_results = [
        linters.CheckResult(
            name="check-name",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result="check-result",
        ),
    ]

    tstamp = datetime.datetime(2020, 2, 1, 15, 40, 33)
    os_platform = OSPlatform(system="SuperUbuntu", release="40.10", machine="SomeRISC")
    with patch("charmcraft.utils.get_os_platform", return_value=os_platform):
        result_filepath = create_manifest(tmp_path, tstamp, bases_config, linting_results)

    assert result_filepath == tmp_path / "manifest.yaml"
    saved = yaml.safe_load(result_filepath.read_text())
    expected = {
        "charmcraft-started-at": "2020-02-01T15:40:33Z",
        "charmcraft-version": __version__,
        "bases": [
            {
                "name": "test-name",
                "channel": "test-channel",
                "architectures": ["arch1"],
            },
            {
                "name": "test-name2",
                "channel": "test-channel2",
                "architectures": ["arch1", "arch2"],
            },
        ],
        "analysis": {
            "attributes": [
                {
                    "name": "check-name",
                    "result": "check-result",
                },
            ],
        },
    }
    assert saved == expected


def test_manifest_no_bases(tmp_path):
    """Manifest without bases (used for bundles)."""
    tstamp = datetime.datetime(2020, 2, 1, 15, 40, 33)
    os_platform = OSPlatform(system="SuperUbuntu", release="40.10", machine="SomeRISC")
    with patch("charmcraft.utils.get_os_platform", return_value=os_platform):
        result_filepath = create_manifest(tmp_path, tstamp, None, [])

    saved = yaml.safe_load(result_filepath.read_text())

    assert result_filepath == tmp_path / "manifest.yaml"
    assert saved == {
        "charmcraft-started-at": "2020-02-01T15:40:33Z",
        "charmcraft-version": __version__,
        "analysis": {"attributes": []},
    }


def test_manifest_checkers_multiple(tmp_path):
    """Multiple checkers, attributes and a linter."""
    linting_results = [
        linters.CheckResult(
            name="attrib-name-1",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result="result-1",
        ),
        linters.CheckResult(
            name="attrib-name-2",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result="result-2",
        ),
        linters.CheckResult(
            name="warning-name",
            check_type=linters.CheckType.LINT,
            url="url",
            text="text",
            result="result",
        ),
    ]

    tstamp = datetime.datetime(2020, 2, 1, 15, 40, 33)
    os_platform = OSPlatform(system="SuperUbuntu", release="40.10", machine="SomeRISC")
    with patch("charmcraft.utils.get_os_platform", return_value=os_platform):
        result_filepath = create_manifest(tmp_path, tstamp, None, linting_results)

    assert result_filepath == tmp_path / "manifest.yaml"
    saved = yaml.safe_load(result_filepath.read_text())
    expected = [
        {
            "name": "attrib-name-1",
            "result": "result-1",
        },
        {
            "name": "attrib-name-2",
            "result": "result-2",
        },
    ]
    assert saved["analysis"]["attributes"] == expected


def test_manifest_image_info_ok(tmp_path, monkeypatch):
    """Include the image info in the manifest."""
    test_image_content = {"some info": ["whatever", 123]}
    monkeypatch.setenv(IMAGE_INFO_ENV_VAR, json.dumps(test_image_content))

    tstamp = datetime.datetime(2020, 2, 1, 15, 40, 33)
    os_platform = OSPlatform(system="SuperUbuntu", release="40.10", machine="SomeRISC")
    with patch("charmcraft.utils.get_os_platform", return_value=os_platform):
        result_filepath = create_manifest(tmp_path, tstamp, None, [])

    saved = yaml.safe_load(result_filepath.read_text())
    assert saved["image-info"] == test_image_content


def test_manifest_image_info_bad(tmp_path, monkeypatch):
    """The format of the image info environment variable is wrong."""
    monkeypatch.setenv(IMAGE_INFO_ENV_VAR, "this is not a json")
    tstamp = datetime.datetime(2020, 2, 1, 15, 40, 33)
    with pytest.raises(CraftError) as cm:
        create_manifest(tmp_path, tstamp, None, [])
    exc = cm.value
    assert str(exc) == "Failed to parse the content of CHARMCRAFT_IMAGE_INFO environment variable"
    assert isinstance(exc.__cause__, json.JSONDecodeError)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Tests for store helpers commands (code in store/charmlibs.py)."""
import hashlib
import pathlib
import sys

import pytest
from craft_cli import CraftError

from charmcraft.utils.charmlibs import (
    collect_charmlib_pydeps,
    get_lib_info,
    get_lib_internals,
    get_libs_from_tree,
    get_name_from_metadata,
)


# region Name-related tests
def test_get_name_from_metadata_ok(tmp_path, monkeypatch):
    """The metadata file is valid yaml, but there is no name."""
    monkeypatch.chdir(tmp_path)

    # put a valid metadata
    metadata_file = tmp_path / "metadata.yaml"
    with metadata_file.open("wb") as fh:
        fh.write(b"name: test-name")

    result = get_name_from_metadata()
    assert result == "test-name"


def test_get_name_from_metadata_no_file(tmp_path, monkeypatch):
    """No metadata file to get info."""
    monkeypatch.chdir(tmp_path)
    result = get_name_from_metadata()
    assert result is None


def test_get_name_from_metadata_bad_content_garbage(tmp_path, monkeypatch):
    """The metadata file is broken."""
    monkeypatch.chdir(tmp_path)

    # put a broken metadata
    metadata_file = tmp_path / "metadata.yaml"
    with metadata_file.open("wb") as fh:
        fh.write(b"\b00\bff -- not a really yaml stuff")

    result = get_name_from_metadata()
    assert result is None


def test_get_name_from_metadata_bad_content_no_name(tmp_path, monkeypatch):
    """The metadata file is valid yaml, but there is no name."""
    monkeypatch.chdir(tmp_path)

    # put a broken metadata
    metadata_file = tmp_path / "metadata.yaml"
    with metadata_file.open("wb") as fh:
        fh.write(b"{}")

    result = get_name_from_metadata()
    assert result is None


# endregion
# region getlibinfo tests


def _create_lib(
    extra_content=None,
    metadata_id=None,
    metadata_api=None,
    metadata_patch=None,
    pydeps=None,
    charm_name="testcharm",
    lib_name="testlib.py",
):
    """Helper to create the structures on disk for a given lib.

    WARNING: this function has the capability of creating INCORRECT structures on disk.
    This is specific for the _get_lib_info tests below, other tests should use the
    functionality provided by the factory.
    """
    base_dir = pathlib.Path("lib")
    lib_file = base_dir / "charms" / charm_name / "v3" / lib_name
    lib_file.parent.mkdir(parents=True, exist_ok=True)

    # save the content to that specific file under custom structure
    if metadata_id is None:
        metadata_id = "LIBID = 'test-lib-id'"
    if metadata_api is None:
        metadata_api = "LIBAPI = 3"
    if metadata_patch is None:
        metadata_patch = "LIBPATCH = 14"

    fields = [metadata_id, metadata_api, metadata_patch]
    if pydeps is not None:
        fields.append(pydeps)
    with lib_file.open("wt", encoding="utf8") as fh:
        for f in fields:
            if f:
                fh.write(f + "\n")
        if extra_content:
            fh.write(extra_content)

    return lib_file


def test_getlibinfo_success_simple(tmp_path, monkeypatch):
    """Simple basic case of success getting info from the library."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib()

    lib_data = get_lib_info(lib_path=test_path)
    assert lib_data.lib_id == "test-lib-id"
    assert lib_data.api == 3
    assert lib_data.patch == 14
    assert lib_data.content_hash is not None
    assert lib_data.content is not None
    assert lib_data.full_name == "charms.testcharm.v3.testlib"
    assert lib_data.path == test_path
    assert lib_data.lib_name == "testlib"
    assert lib_data.charm_name == "testcharm"


@pytest.mark.parametrize(
    "name",
    [
        "charms.testcharm.v3.testlib.py",
        "charms.testcharm.testlib",
        "testcharm.v2.testlib",
        "mycharms.testcharm.v2.testlib",
    ],
)
def test_getlibinfo_bad_name(name):
    """Different combinations of a bad library name."""
    with pytest.raises(CraftError) as err:
        get_lib_info(full_name=name)
    assert str(err.value) == (
        f"Charm library name {name!r} must conform to charms.<charm>.vN.<libname>"
    )


def test_getlibinfo_not_importable_charm_name():
    """Libraries should be save under importable paths."""
    lib_data = get_lib_info(full_name="charms.operator-libs-linux.v0.apt")
    assert lib_data.charm_name == "operator-libs-linux"
    assert lib_data.path == pathlib.Path("lib/charms/operator_libs_linux/v0/apt.py")


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
@pytest.mark.parametrize(
    "path",
    [
        "charms/testcharm/v3/testlib",
        "charms/testcharm/v3/testlib.html",
        "charms/testcharm/v3/testlib.",
        "charms/testcharm/testlib.py",
        "testcharm/v2/testlib.py",
        "mycharms/testcharm/v2/testlib.py",
    ],
)
def test_getlibinfo_bad_path(path):
    """Different combinations of a bad library path."""
    with pytest.raises(CraftError) as err:
        get_lib_info(lib_path=pathlib.Path(path))
    assert str(err.value) == (
        f"Charm library path {path} must conform to lib/charms/<charm>/vN/<libname>.py"
    )


@pytest.mark.parametrize(
    "name",
    [
        "charms.testcharm.v-three.testlib",
        "charms.testcharm.v-3.testlib",
        "charms.testcharm.3.testlib",
        "charms.testcharm.vX.testlib",
    ],
)
def test_getlibinfo_bad_api(name):
    """Different combinations of a bad api in the path/name."""
    with pytest.raises(CraftError) as err:
        get_lib_info(full_name=name)
    assert str(err.value) == (
        "The API version in the library path must be 'vN' where N is an integer."
    )


def test_getlibinfo_missing_library_from_name():
    """Partial case for when the library is not found in disk, starting from the name."""
    test_name = "charms.testcharm.v3.testlib"
    # no create lib!
    lib_data = get_lib_info(full_name=test_name)
    assert lib_data.lib_id is None
    assert lib_data.api == 3
    assert lib_data.patch == -1
    assert lib_data.content_hash is None
    assert lib_data.content is None
    assert lib_data.full_name == test_name
    assert lib_data.path == pathlib.Path("lib") / "charms" / "testcharm" / "v3" / "testlib.py"
    assert lib_data.lib_name == "testlib"
    assert lib_data.charm_name == "testcharm"


def test_getlibinfo_missing_library_from_path():
    """Partial case for when the library is not found in disk, starting from the path."""
    test_path = pathlib.Path("lib") / "charms" / "testcharm" / "v3" / "testlib.py"
    # no create lib!
    lib_data = get_lib_info(lib_path=test_path)
    assert lib_data.lib_id is None
    assert lib_data.api == 3
    assert lib_data.patch == -1
    assert lib_data.content_hash is None
    assert lib_data.content is None
    assert lib_data.full_name == "charms.testcharm.v3.testlib"
    assert lib_data.path == test_path
    assert lib_data.lib_name == "testlib"
    assert lib_data.charm_name == "testcharm"


def test_getlibinfo_metadata_api_different_path_api(tmp_path, monkeypatch):
    """The API value included in the file is different than the one in the path."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(metadata_api="LIBAPI = 99")
    with pytest.raises(CraftError) as err:
        get_lib_info(lib_path=test_path)
    assert str(err.value) == (
        "Library {!r} metadata field LIBAPI is different from the version in the path.".format(
            str(test_path)
        )
    )


# endregion
# region tests for get_lib_internals


def test_getlibinternals_success_simple(tmp_path, monkeypatch):
    """Simple basic case of success getting internals from the library."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib()
    internals = get_lib_internals(test_path)
    assert internals.lib_id == "test-lib-id"
    assert internals.api == 3
    assert internals.patch == 14
    assert internals.pydeps == []
    assert internals.content is not None
    assert internals.content_hash is not None


def test_getlibinternals_success_with_pydeps(tmp_path, monkeypatch):
    """Simple basic successful case that includes pydeps."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(pydeps="PYDEPS = ['foo', 'bar']")
    internals = get_lib_internals(test_path)
    assert internals.lib_id == "test-lib-id"
    assert internals.api == 3
    assert internals.patch == 14
    assert internals.pydeps == ["foo", "bar"]
    assert internals.content is not None
    assert internals.content_hash is not None


def test_getlibinternals_success_content(tmp_path, monkeypatch):
    """Check that content and its hash are ok."""
    extra_content = """
        # extra lines for the file
        # extra non-ascii: ñáéíóú
        # the content is everything, this plus metadata
        # the hash should be of this, excluding metadata
    """
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(extra_content=extra_content)

    internals = get_lib_internals(test_path)
    assert internals.content == test_path.read_text(encoding="utf8")
    assert internals.content_hash == hashlib.sha256(extra_content.encode("utf8")).hexdigest()


def test_getlibinternals_non_toplevel_names(tmp_path, monkeypatch):
    """Test non direct assignments."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(extra_content="logging.getLogger('kazoo.client').disabled = True")
    internals = get_lib_internals(test_path)

    assert internals.lib_id == "test-lib-id"
    assert internals.api == 3
    assert internals.patch == 14
    assert internals.pydeps == []
    assert internals.content is not None
    assert internals.content_hash is not None


def test_getlibinternals_malformed_content(tmp_path, monkeypatch):
    """Some internals field is not really valid."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(extra_content="  broken \n    python  ")
    with pytest.raises(CraftError) as err:
        get_lib_internals(lib_path=test_path)
    assert str(err.value) == rf"Failed to parse Python library {str(test_path)!r}"


@pytest.mark.parametrize(
    ("empty_args", "missing"),
    [
        (["metadata_id"], "LIBID"),
        (["metadata_api"], "LIBAPI"),
        (["metadata_patch"], "LIBPATCH"),
        (["metadata_id", "metadata_api"], "LIBAPI, LIBID"),
        (["metadata_patch", "metadata_api"], "LIBAPI, LIBPATCH"),
        (["metadata_patch", "metadata_id"], "LIBID, LIBPATCH"),
    ],
)
def test_getlibinternals_missing_internals_field(tmp_path, empty_args, missing, monkeypatch):
    """Some internals field is not present."""
    monkeypatch.chdir(tmp_path)
    kwargs = {arg: "" for arg in empty_args}
    test_path = _create_lib(**kwargs)
    with pytest.raises(CraftError) as err:
        get_lib_internals(lib_path=test_path)
    assert str(err.value) == (
        f"Library {str(test_path)!r} is missing the mandatory metadata fields: {missing}."
    )


@pytest.mark.parametrize("value", ["v3", "-3"])
def test_getlibinternals_api_bad_value(tmp_path, value, monkeypatch):
    """The API is not a positive integer."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(metadata_api=f"LIBAPI = {value}")
    with pytest.raises(CraftError) as err:
        get_lib_internals(lib_path=test_path)
    assert str(err.value) == (
        f"Library {str(test_path)!r} metadata field LIBAPI "
        "must be a constant assignment of zero or a positive integer."
    )


@pytest.mark.parametrize("value", ["beta3", "-1"])
def test_getlibinternals_patch_bad_value(tmp_path, value, monkeypatch):
    """The PATCH is not a positive integer."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(metadata_patch="LIBPATCH = {value}")
    with pytest.raises(CraftError) as err:
        get_lib_internals(lib_path=test_path)
    assert str(err.value) == (
        f"Library {str(test_path)!r} metadata field LIBPATCH "
        "must be a constant assignment of zero or a positive integer."
    )


def test_getlibinternals_api_patch_both_zero(tmp_path, monkeypatch):
    """Invalid combination of both API and PATCH being 0."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(metadata_patch="LIBPATCH = 0", metadata_api="LIBAPI = 0")
    with pytest.raises(CraftError) as err:
        get_lib_internals(lib_path=test_path)
    assert str(err.value) == (
        f"Library {str(test_path)!r} metadata fields LIBAPI and LIBPATCH cannot both be zero."
    )


@pytest.mark.parametrize("value", [99, "moño", ""])
def test_getlibinternals_libid_bad_value(tmp_path, value, monkeypatch):
    """The ID is not really a ASCII nonempty string."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(metadata_id=f"LIBID = {value!r}")
    with pytest.raises(CraftError) as err:
        get_lib_internals(lib_path=test_path)
    assert str(err.value) == (
        f"Library {str(test_path)!r} metadata field LIBID "
        "must be a constant assignment of a non-empty ASCII string."
    )


def test_getlibinternals_pydeps_complex(tmp_path, monkeypatch):
    """The PYDEPS field can be multiline, unicode, different quotes."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(
        pydeps="""PYDEPS = [
        'foo',
        "bar",
        "moño",
    ]"""
    )
    internals = get_lib_internals(test_path)
    assert internals.lib_id == "test-lib-id"
    assert internals.api == 3
    assert internals.patch == 14
    assert internals.pydeps == ["foo", "bar", "moño"]
    assert internals.content is not None
    assert internals.content_hash is not None


@pytest.mark.parametrize(
    "value",
    [
        "'foo'",  # a string
        "33",  # other object
        "open()",  # generic code
        "('foo', 'bar')",  # a tuple
        "['foo', 33]",  # a list with wrong fields inside
        "['foo', otherdep]",  # a list with wrong fields inside
    ],
)
def test_getlibinternals_pydeps_bad_value(tmp_path, value, monkeypatch):
    """Different cases with invalid PYDEPS."""
    monkeypatch.chdir(tmp_path)
    test_path = _create_lib(metadata_id=f"PYDEPS = {value}")
    with pytest.raises(CraftError) as err:
        get_lib_internals(lib_path=test_path)
    assert str(err.value) == (
        f"Library {str(test_path)!r} metadata field PYDEPS "
        "must be a constant list of non-empty strings"
    )


# endregion
# region get libs from tree tests


def test_getlibsfromtree_named_currentdir(tmp_path, monkeypatch):
    """Get libs for a specific charm in the current directory."""
    monkeypatch.chdir(tmp_path)
    test_path_1 = _create_lib(charm_name="charm1", lib_name="testlib1.py")
    test_path_2 = _create_lib(charm_name="charm1", lib_name="testlib2.py")
    _create_lib(charm_name="charm2", lib_name="testlib3.py")
    libs_data = get_libs_from_tree(charm_name="charm1")
    assert {data.path for data in libs_data} == {test_path_1, test_path_2}


def test_getlibsfromtree_everything_currentdir(tmp_path, monkeypatch):
    """Get libs for a specific charm in the current directory."""
    monkeypatch.chdir(tmp_path)
    test_path_1 = _create_lib(charm_name="charm1", lib_name="testlib1.py")
    test_path_2 = _create_lib(charm_name="charm1", lib_name="testlib2.py")
    test_path_3 = _create_lib(charm_name="charm2", lib_name="testlib3.py")
    libs_data = get_libs_from_tree()
    assert {data.path for data in libs_data} == {test_path_1, test_path_2, test_path_3}


def test_getlibsfromtree_named_otherdir(tmp_path, monkeypatch):
    """Get libs for a specific charm in other directory."""
    otherdir = tmp_path / "otherdir"
    otherdir.mkdir()
    monkeypatch.chdir(otherdir)
    test_path_1 = _create_lib(charm_name="charm1", lib_name="testlib1.py")
    test_path_2 = _create_lib(charm_name="charm1", lib_name="testlib2.py")
    monkeypatch.chdir(tmp_path)
    _create_lib(charm_name="charm2", lib_name="testlib3.py")
    libs_data = get_libs_from_tree(charm_name="charm1", root=otherdir)
    assert {data.path for data in libs_data} == {test_path_1, test_path_2}


def test_getlibsfromtree_everything_otherdir(tmp_path, monkeypatch):
    """Get libs for a specific charm in other directory."""
    otherdir = tmp_path / "otherdir"
    otherdir.mkdir()
    monkeypatch.chdir(otherdir)
    test_path_1 = _create_lib(charm_name="charm1", lib_name="testlib1.py")
    test_path_2 = _create_lib(charm_name="charm1", lib_name="testlib2.py")
    test_path_3 = _create_lib(charm_name="charm2", lib_name="testlib3.py")
    monkeypatch.chdir(tmp_path)
    libs_data = get_libs_from_tree(root=otherdir)
    assert {data.path for data in libs_data} == {test_path_1, test_path_2, test_path_3}


# endregion
# region pydeps collection tests


def test_collectpydeps_generic(tmp_path, monkeypatch):
    """Collect the PYDEPS from all libs from all charms."""
    otherdir = tmp_path / "otherdir"
    otherdir.mkdir()
    monkeypatch.chdir(otherdir)
    _create_lib(charm_name="charm1", lib_name="lib1.py", pydeps="PYDEPS = ['foo', 'bar']")
    _create_lib(charm_name="charm1", lib_name="lib2.py", pydeps="PYDEPS = ['bar']")
    _create_lib(charm_name="charm2", lib_name="lib3.py")
    _create_lib(charm_name="charm2", lib_name="lib3.py", pydeps="PYDEPS = ['baz']")
    monkeypatch.chdir(tmp_path)
    charmlib_deps = collect_charmlib_pydeps(otherdir)
    assert charmlib_deps == {"foo", "bar", "baz"}


# endregion



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for charm plugin."""
import sys
from unittest.mock import patch

import pydantic
import pytest
from craft_parts import Step

from charmcraft import charm_builder, env, parts

pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Windows not supported")


def test_charmplugin_get_build_package_deb_based(charm_plugin):
    with patch("craft_parts.utils.os_utils.OsRelease.id") as mock_id:
        mock_id.return_value = "ubuntu"

        assert charm_plugin.get_build_packages() == {
            "python3-pip",
            "python3-setuptools",
            "python3-wheel",
            "python3-venv",
            "python3-dev",
        }


def test_charmplugin_get_build_package_yum_based(charm_plugin):
    with patch("craft_parts.utils.os_utils.OsRelease.id") as mock_id:
        mock_id.return_value = "centos"

        assert charm_plugin.get_build_packages() == {
            "autoconf",
            "automake",
            "gcc",
            "gcc-c++",
            "git",
            "make",
            "patch",
            "python3-devel",
            "python3-pip",
            "python3-setuptools",
            "python3-wheel",
        }


def test_charmplugin_get_build_package_centos7(charm_plugin):
    with patch("craft_parts.utils.os_utils.OsRelease.id") as mock_id:
        with patch("craft_parts.utils.os_utils.OsRelease.version_id") as mock_version:
            mock_id.return_value = "centos"
            mock_version.return_value = "7"

            assert charm_plugin.get_build_packages() == {
                "autoconf",
                "automake",
                "gcc",
                "gcc-c++",
                "git",
                "make",
                "patch",
                "rh-python38-python-devel",
                "rh-python38-python-pip",
                "rh-python38-python-setuptools",
                "rh-python38-python-wheel",
            }


def test_charmplugin_get_build_snaps(charm_plugin):
    assert charm_plugin.get_build_snaps() == set()


def test_charmplugin_get_build_environment_ubuntu(charm_plugin, mocker):
    mock_id = mocker.patch("craft_parts.utils.os_utils.OsRelease.id")
    mock_version = mocker.patch("craft_parts.utils.os_utils.OsRelease.version_id")
    mock_id.return_value = "ubuntu"
    mock_version.return_value = "22.04"
    assert charm_plugin.get_build_environment() == {}


def test_charmplugin_get_build_environment_centos_7(charm_plugin, mocker, monkeypatch):
    monkeypatch.setenv("PATH", "/some/path")
    mock_id = mocker.patch("craft_parts.utils.os_utils.OsRelease.id")
    mock_version = mocker.patch("craft_parts.utils.os_utils.OsRelease.version_id")
    mock_id.return_value = "centos"
    mock_version.return_value = "7"
    assert charm_plugin.get_build_environment() == {
        "PATH": "/opt/rh/rh-python38/root/usr/bin:${PATH}"
    }


def test_charmplugin_get_build_commands_ubuntu(charm_plugin, tmp_path, mocker, monkeypatch):
    monkeypatch.setenv("PATH", "/some/path")
    monkeypatch.setenv("SNAP", "snap_value")
    monkeypatch.setenv("SNAP_ARCH", "snap_arch_value")
    monkeypatch.setenv("SNAP_NAME", "snap_name_value")
    monkeypatch.setenv("SNAP_VERSION", "snap_version_value")
    monkeypatch.setenv("http_proxy", "http_proxy_value")
    monkeypatch.setenv("https_proxy", "https_proxy_value")
    monkeypatch.setenv("no_proxy", "no_proxy_value")

    mock_id = mocker.patch("craft_parts.utils.os_utils.OsRelease.id")
    mock_version = mocker.patch("craft_parts.utils.os_utils.OsRelease.version_id")
    mock_register = mocker.patch("craft_parts.callbacks.register_post_step")

    mock_id.return_value = "ubuntu"
    mock_version.return_value = "22.04"

    assert charm_plugin.get_build_commands() == [
        "env -i LANG=C.UTF-8 LC_ALL=C.UTF-8 PATH=/some/path SNAP=snap_value "
        "SNAP_ARCH=snap_arch_value SNAP_NAME=snap_name_value "
        "SNAP_VERSION=snap_version_value http_proxy=http_proxy_value "
        "https_proxy=https_proxy_value no_proxy=no_proxy_value "
        f"{sys.executable} -u -I "
        f"{charm_builder.__file__} "
        f"--builddir {str(tmp_path)}/parts/foo/build "
        f"--installdir {str(tmp_path)}/parts/foo/install "
        f"--entrypoint {str(tmp_path)}/parts/foo/build/entrypoint "
        "-p pip "
        "-p setuptools "
        "-p wheel "
        "-b pkg1 "
        "-b pkg2 "
        "-p pkg3 "
        "-p pkg4 "
        "-r reqs1.txt "
        "-r reqs2.txt"
    ]

    # check the callback is properly registered for running own method after build
    mock_register.assert_called_with(charm_plugin.post_build_callback, step_list=[Step.BUILD])


def test_charmplugin_get_build_commands_centos_7(charm_plugin, tmp_path, mocker, monkeypatch):
    monkeypatch.setenv("PATH", "/some/path")
    monkeypatch.setenv("SNAP", "snap_value")
    monkeypatch.setenv("SNAP_ARCH", "snap_arch_value")
    monkeypatch.setenv("SNAP_NAME", "snap_name_value")
    monkeypatch.setenv("SNAP_VERSION", "snap_version_value")
    monkeypatch.setenv("http_proxy", "http_proxy_value")
    monkeypatch.setenv("https_proxy", "https_proxy_value")
    monkeypatch.setenv("no_proxy", "no_proxy_value")

    mock_id = mocker.patch("craft_parts.utils.os_utils.OsRelease.id")
    mock_version = mocker.patch("craft_parts.utils.os_utils.OsRelease.version_id")
    mock_register = mocker.patch("craft_parts.callbacks.register_post_step")

    mock_id.return_value = "centos"
    mock_version.return_value = "7"

    assert charm_plugin.get_build_commands() == [
        "env -i LANG=C.UTF-8 LC_ALL=C.UTF-8 PATH=/opt/rh/rh-python38/root/usr/bin:/some/path "
        "SNAP=snap_value SNAP_ARCH=snap_arch_value SNAP_NAME=snap_name_value "
        "SNAP_VERSION=snap_version_value http_proxy=http_proxy_value "
        "https_proxy=https_proxy_value no_proxy=no_proxy_value "
        f"{sys.executable} -u -I "
        f"{charm_builder.__file__} "
        f"--builddir {str(tmp_path)}/parts/foo/build "
        f"--installdir {str(tmp_path)}/parts/foo/install "
        f"--entrypoint {str(tmp_path)}/parts/foo/build/entrypoint "
        "-b pip "
        "-b setuptools "
        "-b wheel "
        "-p pip "
        "-p setuptools "
        "-p wheel "
        "-b pkg1 "
        "-b pkg2 "
        "-p pkg3 "
        "-p pkg4 "
        "-r reqs1.txt "
        "-r reqs2.txt"
    ]

    # check the callback is properly registered for running own method after build
    mock_register.assert_called_with(charm_plugin.post_build_callback, step_list=[Step.BUILD])


def test_charmplugin_post_build_metric_collection(charm_plugin):
    with patch("charmcraft.instrum.merge_from") as mock_collection:
        charm_plugin.post_build_callback("test step info")
    mock_collection.assert_called_with(env.get_charm_builder_metrics_path())


def test_charmpluginproperties_invalid_properties():
    content = {"source": ".", "charm-invalid": True}
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.CharmPlugin.properties_class.unmarshal(content)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("charm-invalid",)
    assert err[0]["type"] == "value_error.extra"


def test_charmpluginproperties_entrypoint_ok():
    """Simple valid entrypoint."""
    content = {"source": ".", "charm-entrypoint": "myep.py"}
    properties = parts.CharmPlugin.properties_class.unmarshal(content)
    assert properties.charm_entrypoint == "myep.py"


def test_charmpluginproperties_entrypoint_default():
    """Specific default if not configured."""
    content = {"source": "."}
    properties = parts.CharmPlugin.properties_class.unmarshal(content)
    assert properties.charm_entrypoint == "src/charm.py"


def test_charmpluginproperties_entrypoint_relative(tmp_path):
    """The configuration is stored relative no matter what."""
    absolute_path = tmp_path / "myep.py"
    content = {"source": str(tmp_path), "charm-entrypoint": str(absolute_path)}
    properties = parts.CharmPlugin.properties_class.unmarshal(content)
    assert properties.charm_entrypoint == "myep.py"


def test_charmpluginproperties_entrypoint_outside_project_absolute(tmp_path):
    """The entrypoint must be inside the project."""
    outside_path = tmp_path.parent / "charm.py"
    content = {"source": str(tmp_path), "charm-entrypoint": str(outside_path)}
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.CharmPlugin.properties_class.unmarshal(content)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("charm-entrypoint",)
    assert err[0]["msg"] == f"charm entry point must be inside the project: {str(outside_path)!r}"


def test_charmpluginproperties_entrypoint_outside_project_relative(tmp_path):
    """The entrypoint must be inside the project."""
    outside_path = tmp_path.parent / "charm.py"
    content = {"source": str(tmp_path), "charm-entrypoint": "../charm.py"}
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.CharmPlugin.properties_class.unmarshal(content)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("charm-entrypoint",)
    assert err[0]["msg"] == f"charm entry point must be inside the project: {str(outside_path)!r}"


def test_charmpluginproperties_requirements_default(tmp_path):
    """The configuration is empty by default."""
    content = {"source": str(tmp_path)}
    properties = parts.CharmPlugin.properties_class.unmarshal(content)
    assert properties.charm_requirements == []


def test_charmpluginproperties_requirements_must_exist(tmp_path):
    """The configured files must be present."""
    reqs_path = tmp_path / "reqs.txt"  # not in disk, really
    content = {"source": str(tmp_path), "charm-requirements": [str(reqs_path)]}
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.CharmPlugin.properties_class.unmarshal(content)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("charm-requirements",)
    assert err[0]["msg"] == f"requirements file {str(reqs_path)!r} not found"


def test_charmpluginproperties_requirements_filepresent_ok(tmp_path):
    """If a specific file is present in disk it's used."""
    (tmp_path / "requirements.txt").write_text("somedep")
    content = {"source": str(tmp_path)}
    properties = parts.CharmPlugin.properties_class.unmarshal(content)
    assert properties.charm_requirements == ["requirements.txt"]


def test_charmpluginproperties_requirements_filepresent_but_configured(tmp_path):
    """The specific file is present in disk but configuration takes priority."""
    (tmp_path / "requirements.txt").write_text("somedep")
    (tmp_path / "alternative.txt").write_text("somedep")
    content = {"source": str(tmp_path), "charm-requirements": ["alternative.txt"]}
    properties = parts.CharmPlugin.properties_class.unmarshal(content)
    assert properties.charm_requirements == ["alternative.txt"]



# Copyright 2020-2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""A client to hit the Store."""

import os
import platform
from json.decoder import JSONDecodeError
from typing import Any, Dict

import craft_store
import requests
from craft_cli import CraftError, emit
from craft_store import endpoints
from requests_toolbelt import (  # type: ignore[import]
    MultipartEncoder,
    MultipartEncoderMonitor,
)

from charmcraft import __version__, utils

TESTING_ENV_PREFIXES = ["TRAVIS", "AUTOPKGTEST_TMP"]

ALTERNATE_AUTH_ENV_VAR = "CHARMCRAFT_AUTH"


def build_user_agent():
    """Build the charmcraft's user agent."""
    if any(key.startswith(prefix) for prefix in TESTING_ENV_PREFIXES for key in os.environ):
        testing = " (testing) "
    else:
        testing = " "
    os_platform = "{0.system}/{0.release} ({0.machine})".format(utils.get_os_platform())
    return "charmcraft/{}{}{} python/{}".format(
        __version__, testing, os_platform, platform.python_version()
    )


class AnonymousClient:
    """Lightweight layer that access public store data."""

    def __init__(self, api_base_url: str, storage_base_url: str):
        self.api_base_url = api_base_url.rstrip("/")
        self.storage_base_url = storage_base_url.rstrip("/")
        self._http_client = craft_store.http_client.HTTPClient(user_agent=build_user_agent())

    def request_urlpath_text(self, method: str, urlpath: str, *args, **kwargs) -> str:
        """Return a request.Response to a urlpath."""
        return self._http_client.request(method, self.api_base_url + urlpath, *args, **kwargs).text

    def request_urlpath_json(self, method: str, urlpath: str, *args, **kwargs) -> Dict[str, Any]:
        """Return .json() from a request.Response to a urlpath."""
        response = self._http_client.request(method, self.api_base_url + urlpath, *args, **kwargs)

        try:
            return response.json()
        except JSONDecodeError as json_error:
            raise CraftError(
                f"Could not retrieve json response ({response.status_code} from request"
            ) from json_error


class Client(craft_store.StoreClient):
    """Lightweight layer above StoreClient."""

    def __init__(self, api_base_url: str, storage_base_url: str, ephemeral: bool = False):
        self.api_base_url = api_base_url.rstrip("/")
        self.storage_base_url = storage_base_url.rstrip("/")

        super().__init__(
            base_url=api_base_url,
            storage_base_url=storage_base_url,
            endpoints=endpoints.CHARMHUB,
            application_name="charmcraft",
            user_agent=build_user_agent(),
            environment_auth=ALTERNATE_AUTH_ENV_VAR,
            ephemeral=ephemeral,
        )

    def login(self, *args, **kwargs):
        """Intercept regular login functionality to forbid it when using alternate auth."""
        if os.getenv(ALTERNATE_AUTH_ENV_VAR) is not None:
            raise CraftError(
                f"Cannot login when using alternative auth through {ALTERNATE_AUTH_ENV_VAR} "
                "environment variable."
            )
        return super().login(*args, **kwargs)

    def logout(self, *args, **kwargs):
        """Intercept regular logout functionality to forbid it when using alternate auth."""
        if os.getenv(ALTERNATE_AUTH_ENV_VAR) is not None:
            raise CraftError(
                f"Cannot logout when using alternative auth through {ALTERNATE_AUTH_ENV_VAR} "
                "environment variable."
            )
        return super().logout(*args, **kwargs)

    def request_urlpath_text(self, method: str, urlpath: str, *args, **kwargs) -> str:
        """Return a request.Response to a urlpath."""
        return super().request(method, self.api_base_url + urlpath, *args, **kwargs).text

    def request_urlpath_json(self, method: str, urlpath: str, *args, **kwargs) -> Dict[str, Any]:
        """Return .json() from a request.Response to a urlpath."""
        response = super().request(method, self.api_base_url + urlpath, *args, **kwargs)

        try:
            return response.json()
        except JSONDecodeError as json_error:
            raise CraftError(
                f"Could not retrieve json response ({response.status_code} from request"
            ) from json_error

    def push_file(self, filepath) -> str:
        """Push the bytes from filepath to the Storage."""
        emit.progress(f"Starting to push {str(filepath)!r}")

        with filepath.open("rb") as fh:
            encoder = MultipartEncoder(
                fields={"binary": (filepath.name, fh, "application/octet-stream")}
            )

            # create a monitor (so that progress can be displayed) as call the real pusher
            monitor = MultipartEncoderMonitor(encoder)
            with emit.progress_bar("Uploading...", monitor.len, delta=False) as progress:
                monitor.callback = lambda mon: progress.advance(mon.bytes_read)
                response = self._storage_push(monitor)

        result = response.json()
        if not result["successful"]:
            raise CraftError(f"Server error while pushing file: {result}")

        upload_id = result["upload_id"]
        emit.progress(f"Uploading bytes ended, id {upload_id}")
        return upload_id

    def _storage_push(self, monitor) -> requests.Response:
        """Push bytes to the storage."""
        return super().request(
            "POST",
            self.storage_base_url + "/unscanned-upload/",
            headers={"Content-Type": monitor.content_type, "Accept": "application/json"},
            data=monitor,
        )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem

from charmcraft import parts

FULLY_DEFINED_STRICT_CHARM = {
    "source": ".",
    "plugin": "charm",
    "charm-strict-dependencies": True,
    "charm-binary-python-packages": [],
    "charm-python-packages": [],
    "charm-requirements": [],
    "charm-entrypoint": "src/charm.py",
}
MINIMAL_STRICT_CHARM = {
    "source": ".",
    "plugin": "charm",
    "charm-strict-dependencies": True,
}


@pytest.mark.parametrize(
    ("part_config", "expected"),
    [
        ({}, {"charm-requirements": ["requirements.txt"]}),
        (
            {"charm-requirements": ["requirements.txt"]},
            {"charm-requirements": ["requirements.txt"]},
        ),
        (
            {"charm-requirements": ["requirements.txt"], "charm-binary-python-packages": ["ops"]},
            {"charm-requirements": ["requirements.txt"], "charm-binary-python-packages": ["ops"]},
        ),
    ],
)
def test_partconfig_strict_dependencies_success(fs: FakeFilesystem, part_config, expected):
    """Test various success scenarios for a charm part with strict dependencies."""
    for file in part_config.get("charm-requirements", ["requirements.txt"]):
        fs.create_file(file, contents="ops~=2.5")

    part_config.update(MINIMAL_STRICT_CHARM)
    real_expected = FULLY_DEFINED_STRICT_CHARM.copy()
    real_expected.update(expected)

    actual = parts.process_part_config(part_config)

    assert actual == real_expected


@pytest.mark.parametrize(
    ("part_config", "message"),
    [
        (
            {"charm-requirements": ["req.txt"], "charm-python-packages": ["ops"]},
            "'charm-python-packages' must not be set if 'charm-strict-dependencies' is enabled",
        ),
        (
            {"charm-requirements": ["req.txt"], "charm-binary-python-packages": ["not-here"]},
            "All dependencies must be specified in requirements files for strict dependencies.",
        ),
        (
            {"charm-requirements": ["req.txt"], "charm-binary-python-packages": ["ops>=2.6"]},
            "'charm-binary-python-packages' may contain only package names allowed to be "
            "installed from binary if 'charm-strict-dependencies' is enabled. Invalid "
            "package names: ['ops>=2.6']",
        ),
        ({}, "'charm-strict-dependencies' requires at least one requirements file."),
    ],
)
def test_partconfig_strict_dependencies_failure(fs: FakeFilesystem, part_config, message):
    """Test failure scenarios for a charm part with strict dependencies."""
    for file in part_config.get("charm-requirements", []):
        fs.create_file(file, contents="ops==2.5.1\n")

    part_config.update(MINIMAL_STRICT_CHARM)

    with pytest.raises(Exception) as exc_info:
        parts.process_part_config(part_config)

    assert message in {e["msg"] for e in exc_info.value.errors()}



#!/usr/bin/env python3

# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Setup script for Charmcraft."""

from setuptools import find_packages, setup

from tools.version import determine_version

with open("README.md", encoding="utf8") as fh:
    long_description = fh.read()

install_requires = [
    "craft-cli>=2.3.0",
    "craft-parts>=1.18",
    "craft-providers",
    "craft-store>=2.4",
    "distro>=1.3.0",
    "humanize>=2.6.0",
    "jsonschema",
    "jinja2",
    "pydantic>=1.10,<2.0",
    "python-dateutil",
    "pyyaml",
    "requests",
    "requests-toolbelt",
    "requests-unixsocket",
    "snap-helpers",
    "tabulate",
    # Needed until requests-unixsocket supports urllib3 v2
    # https://github.com/msabramo/requests-unixsocket/pull/69
    "urllib3<2.0",
]

lint_requires = [
    "black>=23.10.1,<24.0.0",
    "codespell[tomli]>=2.2.6,<3.0.0",
    "ruff~=0.1.1",
    "yamllint>=1.32.0,<2.0.0",
]

type_requires = [
    "mypy[reports]~=1.5",
    "pyright==1.1.332",
    "types-python-dateutil",
    "types-requests",
    "types-setuptools",
    "types-tabulate",
    "types-urllib3",
]

dev_requires = [
    "coverage",
    "flake8",
    "pydocstyle",
    "pyfakefs",
    "pytest",
    "pytest-cov",
    "pytest-mock",
    "pytest-check",
    "pytest-subprocess",
    "responses",
    "tox",
]
dev_requires += lint_requires + type_requires

extras_require = {
    "dev": dev_requires,
    "lint": lint_requires,
    "type": type_requires,
}


setup(
    name="charmcraft",
    version=determine_version(),
    author="Facundo Batista",
    author_email="facundo.batista@canonical.com",
    description="The main tool to build, upload, and develop in general the Juju charms.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/canonical/charmcraft",
    license="Apache-2.0",
    packages=find_packages(include=["charmcraft", "charmcraft.*"]),
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    entry_points={
        "console_scripts": ["charmcraft = charmcraft.main:main"],
    },
    python_requires=">=3",
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,  # so we get templates in the wheel
)



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'version' command."""

from craft_cli import emit

from charmcraft import __version__
from charmcraft.cmdbase import BaseCommand

_overview = """
Show charmcraft version.

The output has the following format: `X.Y.Z[+N.gHASH[.dirty]]`

Where:

- `X`, `Y` and `Z` are the major, minor and patch version numbers,
  upgraded when a release is done

- `+N.gHASH` is present if using charmcraft from the project (how many
  commits after last release, and last commit's hash)

- `.dirty` is present if the branch you're executing charmcraft from has
  modifications

Example: `0.3.1+40.g883455b.dirty`
"""


class VersionCommand(BaseCommand):
    """Show the charmcraft version."""

    name = "version"
    help_msg = "Show charmcraft version"
    overview = _overview
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        self.include_format_option(parser)

    def run(self, parsed_args):
        """Run the command."""
        if parsed_args.format:
            info = {"version": __version__}
            emit.message(self.format_content(parsed_args.format, info))
        else:
            emit.message(__version__)



# Copyright 2020-2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Central configuration management.

Using pydantic's BaseModel, this module supports the translation of the
charmcraft.yaml to a python object.

Configuration Schema
====================

type: [string] one of "charm" or "bundle"

charmhub:
  api-url: [HttpUrl] optional, defaults to "https://api.charmhub.io"
  storage-url: [HttpUrl] optional, defaults to "https://storage.snapcraftcontent.com"
  registry-url = [HttpUrl] optional, defaults to "https://registry.jujucharms.com"

parts:
  charm:
    charm-entrypoint: [string] optional, defaults to "src/charm.py"
    charm-requirements: [list of strings] optional, defaults to ["requirements.txt"] if present
    prime: [list of strings]

  bundle:
    prime: [list of strings]

bases: [list of bases and/or long-form base configurations]

analysis:
  ignore:
    attributes: [list of attribute names to ignore]
    linting: [list of linter names to ignore]

actions:
  my-action:
    description: Action as defined at https://juju.is/docs/sdk/actions


Object Definitions
==================

Base
****

Object with the following properties:
- name: [string] name of base
- channel: [string] name of channel
- architectures: [list of strings], defaults to [<host-architecture>]

BaseConfiguration
*****************

Object with the following properties:
- build-on: [list of bases] to build on
- run-on: [list of bases] that build-on entries may run on

"""

import datetime
import pathlib
from typing import Optional

from charmcraft.env import (
    get_managed_environment_project_path,
    is_charmcraft_running_in_managed_mode,
)
from charmcraft.models.charmcraft import CharmcraftConfig, Project
from charmcraft.utils import load_yaml


def load(dirpath: Optional[str]) -> CharmcraftConfig:
    """Load the config from charmcraft.yaml in the indicated directory."""
    if dirpath is None:
        if is_charmcraft_running_in_managed_mode():
            dirpath = get_managed_environment_project_path()
        else:
            dirpath = pathlib.Path.cwd()
    else:
        dirpath = pathlib.Path(dirpath).expanduser().resolve()

    now = datetime.datetime.utcnow()

    content = load_yaml(dirpath / "charmcraft.yaml")
    if content is None:
        # configuration is mandatory only for some commands; when not provided, it will
        # be initialized all with defaults (but marked as not provided for later verification)
        return CharmcraftConfig(
            type="charm",
            project=Project(
                dirpath=dirpath,
                config_provided=False,
                started_at=now,
            ),
            name="missing-charm-name",
            summary="missing-charm-summary",
            description="missing-charm-description",
        )

    return CharmcraftConfig.unmarshal(
        content,
        project=Project(
            dirpath=dirpath,
            config_provided=True,
            started_at=now,
        ),
    )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Extension processor and related utilities."""

from charmcraft.extensions._utils import apply_extensions
from charmcraft.extensions.registry import (
    get_extension_class,
    get_extension_names,
    register,
    unregister,
)

__all__ = [
    "get_extension_class",
    "get_extension_names",
    "apply_extensions",
    "register",
    "unregister",
]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Handlers for manifest.yaml file."""

import datetime
import json
import logging
import os
import pathlib
from typing import Any, Dict, List, Optional

import yaml
from craft_cli import CraftError

import charmcraft.linters
import charmcraft.models.charmcraft
from charmcraft.const import IMAGE_INFO_ENV_VAR

logger = logging.getLogger(__name__)


def create_manifest(
    basedir: pathlib.Path,
    started_at: datetime.datetime,
    bases_config: Optional[charmcraft.models.charmcraft.BasesConfiguration],
    linting_results: List[charmcraft.linters.CheckResult],
) -> pathlib.Path:
    """Create manifest.yaml in basedir for given base configuration.

    For packing bundles, `bases` will be skipped when bases_config is None.
    Charms should always include a valid bases_config.

    :param basedir: Directory to create Charm in.
    :param started_at: Build start time.
    :param bases_config: Relevant bases configuration, if any.

    :returns: Path to created manifest.yaml.
    """
    content: Dict[str, Any] = {
        "charmcraft-version": charmcraft.__version__,
        "charmcraft-started-at": started_at.isoformat() + "Z",
    }

    # Annotate bases only if bases_config is not None.
    if bases_config is not None:
        bases = [
            {
                "name": r.name,
                "channel": r.channel,
                "architectures": r.architectures,
            }
            for r in bases_config.run_on
        ]
        content["bases"] = bases

    # include the linters results (only for attributes)
    attributes_info = [
        {"name": result.name, "result": result.result}
        for result in linting_results
        if result.check_type == charmcraft.linters.CheckType.ATTRIBUTE
    ]
    content["analysis"] = {"attributes": attributes_info}

    # include the image info, if present
    image_info_raw = os.environ.get(IMAGE_INFO_ENV_VAR)
    if image_info_raw:
        try:
            image_info = json.loads(image_info_raw)
        except json.decoder.JSONDecodeError as exc:
            msg = f"Failed to parse the content of {IMAGE_INFO_ENV_VAR} environment variable"
            raise CraftError(msg) from exc
        content["image-info"] = image_info

    filepath = basedir / "manifest.yaml"
    filepath.write_text(yaml.dump(content))
    return filepath



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import pytest

from charmcraft import errors, extensions
from charmcraft.extensions.extension import Extension


class FakeExtension1(Extension):
    """A fake test Extension"""

    name = "fake-extension-1"


class FakeExtension2(Extension):
    """A fake test Extension"""

    name = "fake-extension-2"


class FakeExtension3(Extension):
    """A fake test Extension"""

    name = "fake-extension-3"


@pytest.fixture()
def fake_extensions(stub_extensions):
    for ext_class in (FakeExtension1, FakeExtension2):
        extensions.register(ext_class.name, ext_class)


def test_get_extension_names(fake_extensions):
    assert extensions.get_extension_names() == [
        FakeExtension1.name,
        FakeExtension2.name,
    ]


def test_get_extension_class(fake_extensions):
    assert extensions.get_extension_class(FakeExtension1.name) is FakeExtension1
    assert extensions.get_extension_class(FakeExtension2.name) is FakeExtension2


def test_get_extension_class_error(fake_extensions):
    with pytest.raises(errors.ExtensionError):
        extensions.get_extension_class(FakeExtension3.name)


def test_register(fake_extensions):
    assert FakeExtension3.name not in extensions.get_extension_names()
    extensions.register(FakeExtension3.name, FakeExtension3)
    assert FakeExtension3.name in extensions.get_extension_names()
    assert extensions.get_extension_class(FakeExtension3.name) is FakeExtension3


def test_unregister(fake_extensions):
    assert extensions.get_extension_class(FakeExtension1.name) is FakeExtension1
    extensions.unregister(FakeExtension1.name)
    with pytest.raises(errors.ExtensionError):
        extensions.get_extension_class(FakeExtension1.name)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Utilities related to Python packages."""
import pathlib
import re
import string
import subprocess
from typing import Collection, Iterable, List, Set, Tuple

from charmcraft.errors import MissingDependenciesError

PACKAGE_LINE_REGEX = re.compile(r"^([A-Za-z0-9_.-]+)( *[~<>=!]==?)?")


def get_pypi_packages(*requirements: Iterable[str]) -> Set[str]:
    """Get a set of pypi packages from requirements files.

    :param requirements: An iterable of strings for each requirement.
    :returns: A set of package names and their requirements (e.g. version numbers)
    """
    valid_package_start_chars = string.ascii_letters + string.digits
    packages = set()
    for req in requirements:
        for line in req:
            line = line.strip()
            if not line or line[0] not in valid_package_start_chars:
                continue
            if PACKAGE_LINE_REGEX.match(line):
                packages.add(line)

    return packages


def get_package_names(packages: Iterable[str]) -> Set[str]:
    """Get just the names of packages from an iterable of package lines.

    :param packages: An iterable of package lines (e.g. ["abc==1.0.0", "def"])
    :returns: A set of package names only (e.g. {"abc", "def"})
    """
    names = set()
    for package in packages:
        if match := PACKAGE_LINE_REGEX.match(package):
            names.add(match.group(1))

    return names


def get_requirements_file_package_names(*requirements_files: pathlib.Path) -> Set[str]:
    """Get all the package names from one or more requirements files as a single set."""
    packages = set()
    for file in requirements_files:
        packages |= get_package_names(
            get_pypi_packages(file.read_text().splitlines(keepends=False))
        )
    return packages


def exclude_packages(requirements: Set[str], *, excluded: Collection[str]) -> Set[str]:
    """Filter a set of requirements lines by a collection of package names.

    :param requirements: A set of requirements lines (e.g. {"abc==1.0.0", "def>=1.0"})
    :param excluded: A collection of package names (only) to exclude.
    :returns A filtered set of requirements, excluding the given packages.
    """
    exclusions = set()
    for requirement in requirements:
        if match := PACKAGE_LINE_REGEX.match(requirement):
            if match.group(1) in excluded:
                exclusions.add(requirement)

    return requirements - exclusions


def get_pip_command(
    prefix: Iterable[str],
    requirements_files: Collection[pathlib.Path],
    *,
    source_deps: Collection[str] = (),
    binary_deps: Collection[str] = (),
) -> List[str]:
    """Build a pip command based on requirements files and dependencies.

    :param prefix: The pip command and any earlier arguments.
    :param requirements_files: Paths to the requirements files to include.
    :param source_deps: Additional dependencies that can only be installed from source.
    :param binary_deps: Dependencies (including from requirements files) allowed for binary install.
    :returns: A full pip command line
    """
    charm_packages = get_pypi_packages(source_deps)
    binary_packages = get_pypi_packages(binary_deps)
    requirements_packages = get_requirements_file_package_names(*requirements_files)
    all_packages = charm_packages | binary_packages | requirements_packages
    source_only_packages = sorted(
        get_package_names(all_packages) - get_package_names(binary_packages)
    )

    non_requirements_packages = sorted(
        exclude_packages(
            set(source_deps) | set(binary_deps),
            excluded=get_package_names(requirements_packages),
        )
    )

    if source_only_packages:
        no_binary = [f"--no-binary={','.join(source_only_packages)}"]
    else:
        no_binary = []

    return [
        *prefix,
        *no_binary,
        *(f"--requirement={path}" for path in requirements_files),
        *non_requirements_packages,
    ]


def get_pip_version(pip_cmd: str) -> Tuple[int, ...]:
    """Get the version of pip available from a specific pip command."""
    result = subprocess.run([pip_cmd, "--version"], text=True, capture_output=True, check=True)
    version_data = result.stdout.split(" ")
    if len(version_data) < 2:
        raise ValueError("Unknown pip version")
    version_strings = version_data[1].split(".")
    try:
        return tuple(int(num) for num in version_strings)
    except ValueError:
        raise ValueError(f"Unknown pip version {version_data[1]}")


def validate_strict_dependencies(
    dependencies: Iterable[str], *other_packages: Collection[str]
) -> None:
    """Validate that a charm has an appropriate set of strict dependencies.

    :param dependencies: Packages that are included in known dependencies.
    :param other_packages: Packages from other sources, to ensure these packages are in the
        overall set of dependencies.
    """
    dependency_names = get_package_names(dependencies)

    extra_packages: Set[str] = set()

    for package_set in other_packages:
        other_names = get_package_names(package_set)
        extra_packages |= other_names - dependency_names

    if extra_packages:
        raise MissingDependenciesError(extra_packages)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""YAML-related utilities for Charmcraft."""
from typing import Any, Dict, Optional

import yaml
from craft_cli import emit


def load_yaml(fpath) -> Optional[Dict[str, Any]]:
    """Return the content of a YAML file."""
    if not fpath.is_file():
        emit.debug(f"Couldn't find config file {str(fpath)!r}")
        return None
    try:
        with fpath.open("rb") as fh:
            content = yaml.safe_load(fh)
    except (yaml.error.YAMLError, OSError) as err:
        emit.debug(f"Failed to read/parse config file {str(fpath)!r}: {err!r}")
        return None
    return content



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft Juju Config pydantic model."""

from typing import Any, Dict, Optional

import pydantic

from charmcraft.models.basic import ModelConfigDefaults


class JujuConfig(ModelConfigDefaults):
    """Juju configs for charms.

    See also: https://juju.is/docs/sdk/config
    """

    options: Optional[Dict[str, Dict[str, Any]]]

    @pydantic.validator("options", pre=True)
    def validate_actions(cls, options):
        """Verify options section."""
        if options is None:
            return None
        if not isinstance(options, dict):
            raise ValueError("'options' is not a dictionary")
        for name, option in options.items():
            if not isinstance(option, dict):
                raise ValueError(f"'{name}' is not a dictionary")

            option_keys = set(option.keys())
            if not option_keys.issubset({"description", "type", "default"}):
                invalid_keys = option_keys - {"description", "type", "default"}
                raise ValueError(f"'{name}' has an invalid key(s): {invalid_keys}")

            if "type" not in option:
                raise ValueError(f"'{name}' is missing a type")

            if option["type"] not in ["string", "int", "float", "boolean", "secret"]:
                raise ValueError(
                    f"'{option}' has an invalid type '{option['type']}', "
                    "must be one of: string, int, float, boolean, secret"
                )

        return options



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

from textwrap import dedent

import pytest
from craft_cli import CraftError
from pydantic import AnyHttpUrl
from pydantic.tools import parse_obj_as

from charmcraft.config import load
from charmcraft.metafiles.metadata import parse_charm_metadata_yaml
from charmcraft.models.charmcraft import (
    Base,
    BasesConfiguration,
    Links,
)
from charmcraft.utils import get_host_architecture


def test_load_minimal_metadata_from_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Load a minimal charmcraft.yaml with full metadata. (Spec ST087)"""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel
            """
        )
    )

    config = load(tmp_path)

    assert config.name == "test-charm-name"
    assert config.type == "charm"
    assert config.summary == "test-summary"
    assert config.description == "test-description"
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    )
                ],
                "run-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    )
                ],
            }
        )
    ]
    assert not config.metadata_legacy


def test_load_minimal_metadata_from_charmcraft_yaml_missing_name(
    tmp_path, prepare_charmcraft_yaml
):
    """Load a minimal charmcraft.yaml with metadata. But missing name."""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel
            """
        )
    )

    with pytest.raises(CraftError, match="needs value in field 'name'"):
        load(tmp_path)


def test_load_minimal_metadata_from_charmcraft_yaml_missing_type(
    tmp_path, prepare_charmcraft_yaml
):
    """Load a minimal charmcraft.yaml with metadata. But missing type."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )

    with pytest.raises(CraftError, match="field 'type' required in top-level configuration"):
        load(tmp_path)


def test_load_minimal_metadata_from_charmcraft_yaml_missing_summary(
    tmp_path, prepare_charmcraft_yaml
):
    """Load a minimal charmcraft.yaml with metadata. But missing summary."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            description: test-description

            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )

    with pytest.raises(CraftError, match="needs value in field 'summary'"):
        load(tmp_path)


def test_load_minimal_metadata_from_charmcraft_yaml_missing_description(
    tmp_path, prepare_charmcraft_yaml
):
    """Load a minimal charmcraft.yaml with metadata. But missing description."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary

            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )

    with pytest.raises(CraftError, match="needs value in field 'description'"):
        load(tmp_path)


def test_load_minimal_metadata_from_metadata_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Load a minimal charmcraft.yaml with full metadata. (Spec ST087)"""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm

            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            description: test-description
            """
        ),
    )

    config = load(tmp_path)

    assert config.name == "test-charm-name"
    assert config.type == "charm"
    assert config.summary == "test-summary"
    assert config.description == "test-description"
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    )
                ],
                "run-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    )
                ],
            }
        )
    ]
    assert config.metadata_legacy


def test_load_minimal_metadata_from_metadata_yaml_missing_name(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Load a minimal charmcraft.yaml with metadata.yaml. But missing name."""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm
            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            summary: test-summary
            description: test-description
            """
        ),
    )

    with pytest.raises(CraftError, match="field 'name' required in top-level configuration"):
        load(tmp_path)


def test_load_minimal_metadata_from_metadata_yaml_missing_type(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Load a minimal charmcraft.yaml with metadata.yaml. But missing type."""
    prepare_charmcraft_yaml(
        dedent(
            """
            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            description: test-description
            """
        ),
    )

    with pytest.raises(CraftError, match="field 'type' required in top-level configuration"):
        load(tmp_path)


def test_load_minimal_metadata_from_metadata_yaml_missing_summary(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Load a minimal charmcraft.yaml with metadata.yaml. But missing summary."""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm

            bases:
                - name: test-name
                  channel: test-channel
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            description: test-description
            """
        ),
    )

    with pytest.raises(CraftError, match="field 'summary' required in top-level configuration"):
        load(tmp_path)


def test_load_minimal_metadata_from_metadata_yaml_bad_others_allowed(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Load a minimal charmcraft.yaml with metadata.yaml. Had bad other fields but allowed."""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm

            bases:
                - name: test-name
                  channel: test-channel
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            description: test-description

            peers:
              - aaa
              - bbb
              - ccc
            """
        ),
    )

    load(tmp_path)


def test_load_minimal_metadata_from_metadata_yaml_missing_description(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Load a minimal charmcraft.yaml with metadata.yaml. But missing description."""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm

            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            """
        ),
    )

    with pytest.raises(
        CraftError, match="field 'description' required in top-level configuration"
    ):
        load(tmp_path)


def test_load_full_metadata_from_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Load a charmcraft.yaml with full metadata. (Spec ST087)"""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel

            assumes:
              - test-feature
              - any-of:
                  - extra-feature-1
                  - extra-feature-2
              - all-of:
                  - test-feature-1
                  - test-feature-2

            containers:
              container-1:
                resource: resource-1
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-1
                    location: /var/lib/storage-1
              container-2:
                resource: resource-2
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-2
                    location: /var/lib/storage-2

            devices:
              test-device-1:
                  type: gpu
                  description: gpu
                  countmin: 1
                  countmax: 10

            title: test-title

            extra-bindings:
              test-binding-1: binding-1

            links:
              issues: https://example.com/issues
              contact:
                - https://example.com/contact
                - contact@example.com
                - "IRC #example"
              documentation: https://example.com/docs
              source:
                - https://example.com/source
                - https://example.com/source2
                - https://example.com/source3
              website:
                - https://example.com/

            peers:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            provides:
              provide-1:
                interface: eth1
                limit: 1
                optional: true
                scope: global

            requires:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            resources:
              resource-1:
                type: file
                description: resource-1
                filename: /path/to/resource-1

            storage:
              storage-1:
                type: filesystem
                description: storage-1
                location: /var/lib/storage-1
                shared: true
                read-only: false
                multiple: 5G
                minimum-size: 5G
                properties:
                    - transient

            subordinate: true

            terms:
              - https://example.com/terms
              - https://example.com/terms2
            """
        ),
    )

    config = load(tmp_path)
    config_dict = config.dict()

    # remove unrelated keys. but they should exist in the config

    del config_dict["actions"]
    del config_dict["analysis"]
    del config_dict["charmhub"]
    del config_dict["config"]
    del config_dict["project"]
    del config_dict["parts"]

    assert config_dict == {
        "name": "test-charm-name",
        "type": "charm",
        "summary": "test-summary",
        "description": "test-description",
        "bases": [
            BasesConfiguration(
                **{
                    "build-on": [
                        Base(
                            name="test-name",
                            channel="test-channel",
                            architectures=[get_host_architecture()],
                        )
                    ],
                    "run-on": [
                        Base(
                            name="test-name",
                            channel="test-channel",
                            architectures=[get_host_architecture()],
                        )
                    ],
                }
            )
        ],
        "assumes": [
            "test-feature",
            {"any-of": ["extra-feature-1", "extra-feature-2"]},
            {"all-of": ["test-feature-1", "test-feature-2"]},
        ],
        "containers": {
            "container-1": {
                "resource": "resource-1",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-1", "location": "/var/lib/storage-1"}],
            },
            "container-2": {
                "resource": "resource-2",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-2", "location": "/var/lib/storage-2"}],
            },
        },
        "devices": {
            "test-device-1": {"type": "gpu", "description": "gpu", "countmin": 1, "countmax": 10}
        },
        "title": "test-title",
        "peers": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "provides": {
            "provide-1": {"interface": "eth1", "limit": 1, "optional": True, "scope": "global"}
        },
        "requires": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "resources": {
            "resource-1": {
                "type": "file",
                "description": "resource-1",
                "filename": "/path/to/resource-1",
            },
        },
        "storage": {
            "storage-1": {
                "type": "filesystem",
                "description": "storage-1",
                "location": "/var/lib/storage-1",
                "shared": True,
                "read-only": False,
                "multiple": "5G",
                "minimum-size": "5G",
                "properties": ["transient"],
            }
        },
        "subordinate": True,
        "terms": ["https://example.com/terms", "https://example.com/terms2"],
        "extra_bindings": {"test-binding-1": "binding-1"},
        "links": Links(
            contact=["https://example.com/contact", "contact@example.com", "IRC #example"],
            documentation=parse_obj_as(AnyHttpUrl, "https://example.com/docs"),
            issues=parse_obj_as(AnyHttpUrl, "https://example.com/issues"),
            source=[
                parse_obj_as(AnyHttpUrl, "https://example.com/source"),
                parse_obj_as(AnyHttpUrl, "https://example.com/source2"),
                parse_obj_as(AnyHttpUrl, "https://example.com/source3"),
            ],
            website=[parse_obj_as(AnyHttpUrl, "https://example.com/")],
        ),
        "metadata_legacy": False,
    }


def test_load_full_metadata_from_metadata_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml
):
    """Load a charmcraft.yaml with full metadata.yaml. (Legacy)"""
    prepare_charmcraft_yaml(
        dedent(
            """
            type: charm
            bases:
              - name: test-name
                channel: test-channel
            """
        ),
    )
    prepare_metadata_yaml(
        dedent(
            """
            name: test-charm-name
            summary: test-summary
            description: test-description
            assumes:
              - test-feature
              - any-of:
                  - extra-feature-1
                  - extra-feature-2
              - all-of:
                  - test-feature-1
                  - test-feature-2

            containers:
              container-1:
                resource: resource-1
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-1
                    location: /var/lib/storage-1
              container-2:
                resource: resource-2
                bases:
                  - name: ubuntu
                    channel: 22.04
                    architectures:
                      - x86_64
                mounts:
                  - storage: storage-2
                    location: /var/lib/storage-2

            devices:
              test-device-1:
                  type: gpu
                  description: gpu
                  countmin: 1
                  countmax: 10

            display-name: test-title

            docs: https://example.com/docs

            extra-bindings:
              test-binding-1: binding-1

            issues: https://example.com/issues

            maintainers:
              - https://example.com/contact
              - contact@example.com
              - "IRC #example"

            peers:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            provides:
              provide-1:
                interface: eth1
                limit: 1
                optional: true
                scope: global

            requires:
              peer-1:
                interface: eth0
                limit: 1
                optional: true
                scope: global

            resources:
              resource-1:
                type: file
                description: resource-1
                filename: /path/to/resource-1

            source:
              - https://example.com/source
              - https://example.com/source2
              - https://example.com/source3

            storage:
              storage-1:
                type: filesystem
                description: storage-1
                location: /var/lib/storage-1
                shared: true
                read-only: false
                multiple: 5G
                minimum-size: 5G
                properties:
                  - transient

            subordinate: true

            terms:
              - https://example.com/terms
              - https://example.com/terms2

            website:
              - https://example.com/
            """
        ),
    )

    config = load(tmp_path)
    metadata = parse_charm_metadata_yaml(tmp_path)

    assert config.name == "test-charm-name"
    assert config.type == "charm"
    assert config.summary == "test-summary"
    assert config.description == "test-description"
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    )
                ],
                "run-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    )
                ],
            }
        )
    ]
    assert config.metadata_legacy

    metadata_dict = metadata.dict()
    assert metadata_dict == {
        "name": "test-charm-name",
        "summary": "test-summary",
        "description": "test-description",
        "assumes": [
            "test-feature",
            {"any-of": ["extra-feature-1", "extra-feature-2"]},
            {"all-of": ["test-feature-1", "test-feature-2"]},
        ],
        "containers": {
            "container-1": {
                "resource": "resource-1",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-1", "location": "/var/lib/storage-1"}],
            },
            "container-2": {
                "resource": "resource-2",
                "bases": [{"name": "ubuntu", "channel": 22.04, "architectures": ["x86_64"]}],
                "mounts": [{"storage": "storage-2", "location": "/var/lib/storage-2"}],
            },
        },
        "devices": {
            "test-device-1": {"type": "gpu", "description": "gpu", "countmin": 1, "countmax": 10}
        },
        "display_name": "test-title",
        "peers": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "provides": {
            "provide-1": {"interface": "eth1", "limit": 1, "optional": True, "scope": "global"}
        },
        "requires": {
            "peer-1": {"interface": "eth0", "limit": 1, "optional": True, "scope": "global"}
        },
        "resources": {
            "resource-1": {
                "type": "file",
                "description": "resource-1",
                "filename": "/path/to/resource-1",
            },
        },
        "storage": {
            "storage-1": {
                "type": "filesystem",
                "description": "storage-1",
                "location": "/var/lib/storage-1",
                "shared": True,
                "read-only": False,
                "multiple": "5G",
                "minimum-size": "5G",
                "properties": ["transient"],
            }
        },
        "subordinate": True,
        "terms": ["https://example.com/terms", "https://example.com/terms2"],
        "extra_bindings": {"test-binding-1": "binding-1"},
        "docs": parse_obj_as(AnyHttpUrl, "https://example.com/docs"),
        "issues": parse_obj_as(AnyHttpUrl, "https://example.com/issues"),
        "maintainers": ["https://example.com/contact", "contact@example.com", "IRC #example"],
        "source": [
            parse_obj_as(AnyHttpUrl, "https://example.com/source"),
            parse_obj_as(AnyHttpUrl, "https://example.com/source2"),
            parse_obj_as(AnyHttpUrl, "https://example.com/source3"),
        ],
        "website": [parse_obj_as(AnyHttpUrl, "https://example.com/")],
    }


def test_load_full_actions_in_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Load a charmcraft.yaml with full actions."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel

            actions:
              pause:
                description: Pause the database.
              resume:
                description: Resume a paused database.
              snapshot:
                description: Take a snapshot of the database.
                params:
                  filename:
                    type: string
                    description: The name of the snapshot file.
                  compression:
                    type: object
                    description: The type of compression to use.
                    properties:
                      kind:
                        type: string
                        enum: [gzip, bzip2, xz]
                      quality:
                        description: Compression quality
                        type: integer
                        minimum: 0
                        maximum: 9
                required: [filename]
                additionalProperties: false
            """
        )
    )

    config = load(tmp_path)

    assert config.actions.dict(include={"actions"}, exclude_none=True, by_alias=True)[
        "actions"
    ] == {
        "pause": {"description": "Pause the database."},
        "resume": {"description": "Resume a paused database."},
        "snapshot": {
            "description": "Take a snapshot of the database.",
            "params": {
                "filename": {
                    "type": "string",
                    "description": "The name of the snapshot file.",
                },
                "compression": {
                    "type": "object",
                    "description": "The type of compression to use.",
                    "properties": {
                        "kind": {"type": "string", "enum": ["gzip", "bzip2", "xz"]},
                        "quality": {
                            "description": "Compression quality",
                            "type": "integer",
                            "minimum": 0,
                            "maximum": 9,
                        },
                    },
                },
            },
            "required": ["filename"],
            "additionalProperties": False,
        },
    }


def test_load_full_actions_in_actions_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_actions_yaml
):
    """Load a charmcraft.yaml with full actions."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description
            """
        ),
    )

    prepare_actions_yaml(
        dedent(
            """
            pause:
              description: Pause the database.
            resume:
              description: Resume a paused database.
            snapshot:
              description: Take a snapshot of the database.
              params:
                filename:
                  type: string
                  description: The name of the snapshot file.
                compression:
                  type: object
                  description: The type of compression to use.
                  properties:
                    kind:
                      type: string
                      enum: [gzip, bzip2, xz]
                    quality:
                      description: Compression quality
                      type: integer
                      minimum: 0
                      maximum: 9
              required: [filename]
              additionalProperties: false
            """
        ),
    )

    config = load(tmp_path)

    assert config.actions.dict(include={"actions"}, exclude_none=True, by_alias=True)[
        "actions"
    ] == {
        "pause": {"description": "Pause the database."},
        "resume": {"description": "Resume a paused database."},
        "snapshot": {
            "description": "Take a snapshot of the database.",
            "params": {
                "filename": {
                    "type": "string",
                    "description": "The name of the snapshot file.",
                },
                "compression": {
                    "type": "object",
                    "description": "The type of compression to use.",
                    "properties": {
                        "kind": {"type": "string", "enum": ["gzip", "bzip2", "xz"]},
                        "quality": {
                            "description": "Compression quality",
                            "type": "integer",
                            "minimum": 0,
                            "maximum": 9,
                        },
                    },
                },
            },
            "required": ["filename"],
            "additionalProperties": False,
        },
    }


@pytest.mark.parametrize(
    "bad_name",
    [
        "is",
        "-snapshot",
        "111snapshot",
    ],
)
def test_load_bad_actions_in_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml, bad_name):
    """Load a bad actions in charmcraft.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            f"""
            name: test-charm-name
            type: charm
            bases:
              - name: test-name
                channel: test-channel
            actions:
              pause:
                description: Pause the database.
              resume:
                description: Resume a paused database.
              {bad_name}:
                description: Take a snapshot of the database.
            """
        )
    )

    with pytest.raises(CraftError):
        load(tmp_path)


@pytest.mark.parametrize(
    "bad_name",
    [
        "is",
        "-snapshot",
        "111snapshot",
    ],
)
def test_load_bad_actions_in_actions_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_actions_yaml, bad_name
):
    """Load a bad actions in actions.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            bases:
              - name: test-name
                channel: test-channel
            """
        )
    )
    prepare_actions_yaml(
        dedent(
            f"""\
            actions:
              pause:
                description: Pause the database.
              resume:
                description: Resume a paused database.
              {bad_name}:
                description: Take a snapshot of the database.
            """
        )
    )

    with pytest.raises(CraftError):
        load(tmp_path)


def test_load_actions_in_charmcraft_yaml_and_actions_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_actions_yaml
):
    """Load actions in charmcraft.yaml and actions.yaml at the same time."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel

            actions:
              pause:
                description: Pause the database.
              resume:
                description: Resume a paused database.
              snapshot:
                description: Take a snapshot of the database.
                params:
                  filename:
                    type: string
                    description: The name of the snapshot file.
                  compression:
                    type: object
                    description: The type of compression to use.
                    properties:
                      kind:
                        type: string
                        enum: [gzip, bzip2, xz]
                      quality:
                        description: Compression quality
                        type: integer
                        minimum: 0
                        maximum: 9
                required: [filename]
                additionalProperties: false
            """
        ),
    )
    prepare_actions_yaml(
        dedent(
            """
            pause:
              description: Pause the database.
            """
        ),
    )

    msg = (
        "'actions.yaml' file not allowed when an 'actions' section "
        "is defined in 'charmcraft.yaml' in field 'actions'"
    )

    with pytest.raises(CraftError, match=msg):
        load(tmp_path)


def test_load_config_in_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Load a config in charmcraft.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            bases:
              - name: test-name
                channel: test-channel

            config:
              options:
                test-int:
                  default: 123
                  description: test-1
                  type: int
                test-string:
                  description: test-2
                  type: string
                test-float:
                  default: 1.23
                  type: float
                test-bool:
                  default: true
                  type: boolean
                test-secret:
                  type: secret
            """
        )
    )
    config = load(tmp_path)

    assert config.config.dict(include={"options"}, by_alias=True) == {
        "options": {
            "test-int": {"default": 123, "description": "test-1", "type": "int"},
            "test-string": {"description": "test-2", "type": "string"},
            "test-float": {"default": 1.23, "type": "float"},
            "test-bool": {"default": True, "type": "boolean"},
            "test-secret": {"type": "secret"},
        },
    }


def test_load_config_in_config_yaml(tmp_path, prepare_charmcraft_yaml, prepare_config_yaml):
    """Load a config in config.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description
            """
        ),
    )
    prepare_config_yaml(
        dedent(
            """
            options:
              test-int:
                default: 123
                description: test-1
                type: int
              test-string:
                description: test-2
                type: string
              test-float:
                default: 1.23
                type: float
              test-bool:
                default: true
                type: boolean
              test-secret:
                type: secret
            """
        ),
    )
    config = load(tmp_path)

    assert config.config.dict(include={"options"}, by_alias=True) == {
        "options": {
            "test-int": {"default": 123, "description": "test-1", "type": "int"},
            "test-string": {"description": "test-2", "type": "string"},
            "test-float": {"default": 1.23, "type": "float"},
            "test-bool": {"default": True, "type": "boolean"},
            "test-secret": {"type": "secret"},
        },
    }


@pytest.mark.parametrize(
    "config_yaml",
    [
        "",
        "options:",
        "options:\n  test-int:",
        "options: not a dict",
    ],
)
def test_load_bad_config_in_config_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_config_yaml, config_yaml
):
    """Load a bad config in config.yaml. Should not raise an error since check unenforced."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description
            """
        ),
    )
    prepare_config_yaml(config_yaml)
    config = load(tmp_path)

    assert config.config is None


def test_load_bad_config_in_charmcraft_yaml(tmp_path, prepare_charmcraft_yaml):
    """Load a config in charmcraft.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """
            name: test-charm-name
            type: charm
            summary: test-summary
            description: test-description

            config:
              options:
                test-int:
                  default: 123
                  descriptionn: test-1
                  type: int
                test-string:
                  description: test-2
                  type: string
                test-float:
                  default: 1.23
                  type: float
                test-bool:
                  default: true
                  type: boolean
                test-secret:
                  type: secret
            """
        )
    )

    with pytest.raises(
        CraftError,
        match=r"'test-int' has an invalid key\(s\): {'descriptionn'} in field 'config.options'",
    ):
        load(tmp_path)
